INCLUDE "gbhw.inc"
INCLUDE "utility.inc"
INCLUDE "irq.inc"
INCLUDE "ibmpc1.inc"
INCLUDE "thread.inc"

SECTION "Thread",CODE

;*******************************************************************
;
; Thread + Timer Interrupt - This must be in BANK0
;
; (GnU) james@ebom.com.au 1/12/1998
;
;
; Createthread	- Create a new thread
;
; Sleep	        - Hands over control to the next thread or if 
;		- all threads are sleeping, halts the gameboy 
;		- to save power
;
;        *note* Sleep is used only if USESLEEP is defined.
;
;        Sleep will crash if you leave your gameboy running for
;        longer than 17 days because of the way I am doing the
;        sleep scheduling. Its a small price to pay for avoiding
;        a lot of code in the speed sensitive context switching
;        ISR.
;
; Own	        - Sleeps till the resource is free and then takes 
;		- ownership of a resource
; Disown        - Frees up a resource
;
; Relenquish    - Hands the CPU over to the next thread in the ring.
;
;
; Note - Each thread is a virtual z80 on its own. But you can't
;        rely on the unused part of your stack that is
;	 even though you pop a value off the stack the data is still
;        there and is re-writen only when you push something over it
;	 because the context switcher uses the local stack
;        this will only effect you if you are certifiably mad and
;	 do such evil as
;
;	 pop de
;	 ; something else
;	 add SP,-2
;	 pop de        <-  This may not be the same value of de 
;                          you had previously. Normally (Unthreaded) 
;			   it would be.
;
;
;
;
;******************************************************************
; Changes
; 
; 3/12/1998  Added CreateThread	james@ebom.org
; 4/12/1998  Added Sleep	james@ebom.org
; 6/12/1998  Added Own/Disown	james@ebom.org
; 8/12/1998  Added Relenquish	james@ebom.org




;
; This will init some critical.
;
; [x] Tested
; [ ] Impropper Tested
;

InitThreadLibrary::
	push af
	; set clock = 00-00-00-01
	xor 	a

	ld	[clock_high],a
	ld	[clock_high+1],a
	ld	[clock_low],a
	inc	a
	ld	[clock_low+1],a
	pop  af
	ret

;------------------------------------------------
; Owns a resource
;
; [ ] Tested
; [ ] Impropper Tested
; [ ] Dynamic Tested
;
; corrupts af
;------------------------------------------------

Own:: ; hl = address of semaphore flag

.slickwait
	di
	ld	a,[hl]
	and	$ff
	jr	z,.its_ours
	ei
	ld	de,$ff

	call	Relenquish	
	jr	.slickwait


.its_ours

	; its ours
	inc	a
	ld	[hl],a

	ei
	ret


;------------------------------------------------
; Disowns a resource - assumes you actually own 
; it.
;
; [ ] Tested
; [ ] Impropper Tested
; [ ] Dynamic Tested
;
; corrupts af
;------------------------------------------------

Disown:: ; hl = address of semaphore flag


	di
	ld	a,[hl]
	dec	a
	ld	[hl],a
	ei

	ret


;------------------------------------------------
;
; Sleep(de) 
;
; This puts a thread to sleep for de timer ticks
;
; [x] Tested
; [ ] Impropper Tested
; [x] Dynamic Tested
;
;
; corrupts af,bc,de,hl
;------------------------------------------------

IF      DEF(USESLEEP)

Sleep:: ; de = clicks to sleep for

	di

        ld      hl,context
        ld      a,[hl+]
        ld      b,a
        ld      a,[hl+]
        ld      c,a

	ld	h,b
	ld	l,c

	; first we make context->sleep_until = be

	push	bc
	ld	b,0
	ld	c,context_sleeptill_high
	add	hl,bc
	pop	bc

	xor	a
	; a =  low.low 
	ld	[hl+],a
	ld	[hl+],a
	ld	a,d
	ld	[hl+],a
	ld	a,e
	ld	[hl+],a

	; now we add the current time to this handle to get the real time
	; we should sleep until

	ld	h,b
	ld	l,c

	; point at low.low
	ld	b,0
	ld	c,context_sleeptill_low+1

	add	hl,bc
	ld	a,[hl]

	; a = clock.low.low (a) + context.low.low (d)
	ld	d,a
	ld	a,[clock_low+1]
	add	d

	; put a back
	ld	[hl-],a

	; get the next one
	ld	a,[hl]

	; Do we need to cary..
	jr	nc,.nocarry1

	; carry
	inc	a

.nocarry1

	; so a now contains low.high

	; a = clock.low.high (a) + context.low.high (d)
	ld	d,a
	ld	a,[clock_low]
	add	d

	; put a back
	ld	[hl-],a

	; get the next one
	ld	a,[hl]

	; Do we need to cary..
	jr	nc,.nocarry2

	; carry
	inc	a

.nocarry2

	; so a now contains low.high

	; a = clock.high.low (a) + context.high.low (d)
	ld	d,a
	ld	a,[clock_high+1]
        add     d
        
        ; put a back
        ld      [hl-],a

	; get the next one
	ld	a,[hl]
        
        ; Do we need to cary..
        jr      nc,.nocarry3
        
        ; carry
        inc     a
        
.nocarry3
        
        ; so a now contains high.high

	; a = clock.high.high (a) + context.high.high (d)
        ld      d,a
        ld      a,[clock_high]
        add     d
        
        ; put a back
        ld      [hl-],a


	; get the next one
	ld	a,[hl]
        
	; if we need to carry then the gameboy has been on for 17days straight and
	; could probably do with a good crash.. :)

	ei

 	call Relenquish

	ret
ENDC

;----------------------------------------------
;
; CreateThread(bc,de)
;
; This creates a thread and adds it to the ring.
;
; [x] Tested
; [ ] Impropper Tested
; [ ] Dynamic Tested
;
;----------------------------------------------

CreateThread:: ; bc = code de = contextblock

	; we save all the registers
	pusha

	;
	; Now we add de to the list bu updating
	; head and tail if its the first one in the list
	; or just tail if its being added to the end
		
	; we also make sure that head points to thiscontext.next
	; and that thiscontext.tail.next points to thiscontext

	; keep bc safe - quickly

	push	bc
	; load the current context and check if its null
        ld      hl,context
        ld      a,[hl+]
        ld      b,a
        ld      a,[hl]
        or	b

	pop	bc	; pop does nothing to the flags

	; if our context pointer is null then set it to this thread
	; bacause it means its our first and only thread

        jr	nz,.context_wasnt_null	

	; context = de
        ld      hl,context
        ld      a,d
        ld      [hl+],a   
        ld      a,e
        ld      [hl],a

	; ring.head = de
	ld      hl,ringhead
        ld      a,d
        ld      [hl+],a   
        ld      a,e
        ld      [hl],a

	jr	.context_was_null

.context_wasnt_null	; or you will die

	;
	; We are adding a second
	;

	; Point the previous context to this one

	push	bc

	; load the current context and check if its null

        ld      hl,ringtail
        ld      a,[hl+]
        ld      b,a
        ld      a,[hl]
        ld	c,a

	; bc now contains the previous context
	;
	; ring.tail->next = be
	
        ld      h,b
        ld      l,c
        ld      a,d
        ld      [hl+],a   
        ld      a,e
        ld      [hl],a
	
	pop	bc	; pop does nothing to the flags

.context_was_null	

	; ring.tail = de

	ld      hl,ringtail
        ld      a,d
        ld      [hl+],a   
        ld      a,e
        ld      [hl],a

	; this.next = first context in chain

	push	bc

        ld      hl,context
        ld      a,[hl+]
        ld      b,a
        ld      a,[hl]
	ld	c,a

	; bc = context

	ld	h,d
	ld	l,e

	; this.next = bc

        ld      a,b
        ld      [hl+],a   
        ld      a,c
        ld      [hl+],a

	xor	a

	; this.sleepfor_high = 0
        ld      [hl+],a   
        ld      [hl+],a   
	; this.sleepfor_low  = 0
        ld      [hl+],a   
        ld      [hl+],a   

	pop 	bc

	; bc = location of executable code
	; This can be anywhere - no restrictions

	; de = location of contextblock sizeof contextlength (currently 128 bytes)
	; in the future I will add support to vary the size of the context block
	; that will probably be passed in hl but for the moment it assumes that
	; the block size in 128. Blame me.

        ; contextblock.pc = thread1loop
        
        ; move our PC onto our fake interrupt stack push in reverse byte order
	        
        ld      hl,contextblocksize - 2
	add	hl,de
        ld      a,c
        ld      [hl+],a
        ld      a,b
        ld      [hl],a
        
        ; put the threads pc in pre-rolled
        ; stack frame as a return address
	; de = location od the context bloxk        
	; What we are doing here is creating
	; a fake area of stack that has all 
	; the registers and a return address
	; so that after we repoint the sp
	; and go reti - It will like we are
	; returning from an interrupt.

	ld 	hl,contextblocksize - 10
	add	hl,de
	ld	b,h;
	ld	c,l;
	; bc now contains the final resting space of the sp
	
	; save the location of the sp buffer
        ld      hl,context_sp
	add	hl,de
        ld      a,b
        ld      [hl+],a
        ld      a,c
        ld      [hl],a
	popa
	ret

;----------------------------------------------
;
; Relenquish()
;
; This gives up the cpu to the next context in the ring		
;
; [x] Tested
; [ ] Impropper Tested
; [ ] Dynamic Tested
;
; corrupts bc
;
;----------------------------------------------		
Relenquish::

;
; This could all be replaced if we could generate a timer interrupt immediately.
;
;

	di

	ld	bc,endofthetunnel

	push	bc
        push    af
        push    bc
        push    de
        push    hl

	; so now our registers and our fake return address are saved

        ;hl = *context
        ld      hl,context
        ld      a,[hl+]
        ld      b,a
        ld      a,[hl]
        ld      c,a	
	ld	h,b
	ld	l,c
	
	; bc = next
	ld	a,[hl+]
	ld	b,a
	ld	a,[hl+]
	ld	c,a
	
        ; save the current stack to hl
        ld      [tempsp],sp
        ld      a,[tempsp+1]
        ld      [hl+],a
        ld      a,[tempsp]
        ld      [hl+],a

	; context = bc

        ld      hl,context
	ld	a,b
	ld	[hl+],a
	ld	a,c
	ld	[hl+],a

	ld	h,b
	ld	l,c

	inc	hl
	inc	hl

	ld	a,[hl+]
	ld	b,a	
	ld	a,[hl+]
	ld	c,a	

	ld	h,b
	ld	l,c
	
	; load new stack pointer

	ld	sp,hl

        pop	hl
	pop	de
	pop	bc
	pop	af
	
	ei

	ret	; which should take us to where we were before with al registers preserved

endofthetunnel:	; this is where we will pop out
	
	ret

;*******************************************************************
;*
;* irq_Init - Initialize interrupt system
;*
;*******************************************************************

irq_Init::
        xor     a,a
        ld      [_HW+rIF],a
        ld      [_HW+rIE],a
	ei
        ret

StartThreadEngine::
        ld      a,[_HW+rIE]
	or	IEF_TIMER
	ld      [_HW+rIE],a        
        ret

; --    
; -- Internal interrupt handling routines
; --
        

irq_LCDC::
        ei
        reti


irq_Timer::
	;------------------------
	;
	; Context switch	
	;
	;------------------------

	; pc is on the stack
	;
	; 1 We save the current sp
	; 2 Change the context pointer
	;   to point to the next context
	;   in the ring
	; 3 load the new sp so we are pointing
	;   at the new contexts previous stack space.
	; 4 Retrieve all the previous registers off
	;   the new stack
	; 5 'reti' and end up at the new stack
   	
	; save current registers into this stack

        push	af

	ld	a,[tmodulus]
	inc	a

	cp	a,4

	jr	z,.tmc

	ld	[tmodulus],a
	pop	af
	reti

.tmc
	xor	a
	ld	[tmodulus],a


	push	bc
	push	de
	push	hl

IF      DEF(USESLEEP)

        ; Increment the clock   
        ld      hl,clock_low+1
        ld      a,[hl]
        inc     a
        ld      [hl-],a
        ld      a,[hl]
	jr	nz,.noccarry1
	inc 	a
.noccarry1
        ld      [hl-],a
        ld      a,[hl]
	jr	nz,.noccarry2
	inc 	a

;	call	update0
;	call	update1
;	call	update2
;	call	update3
;	call	update4
;
;.profile	
;	jr	.profile

.noccarry2
        ld      [hl-],a
        ld      a,[hl]
	jr	nz,.noccarry3
	inc 	a
.noccarry3
        ld      [hl],a

	; should test for overflow here. but it will only
	; happen every 17 days

ENDC


	; - profile code
        ;pop	hl
	;pop	de
	;pop	bc
	;pop	af
        ;reti  
	; profile code
	
        ;---------------------------------------------------------
        ; 
        ; now we start loading the new context
        ;
        
        ;bc = *context
 	        
        ld      hl,context
        ld      a,[hl+]
        ld      b,a   
        ld      a,[hl]
        ld      c,a
        
	; copy to hl 
        ld      h,b
        ld      l,c 

	;hl = bc

        ; hl now points at our current context of which the first two bytes is
	; the address of the next context in this chain
        
	; bc = *context.next

        ld      a,[hl+]
        ld      b,a
        ld      a,[hl+]
        ld      c,a 
      
	; hl now points to the context stackpointer
	
	; save the current stack to hl
 
        ld      [tempsp],sp
        ld      a,[tempsp+1]
        ld      [hl+],a
        ld      a,[tempsp]
        ld      [hl+],a   

        ; so now we can switch contexts
        
.switch

        ld      hl,context
        ld      a,b
        ld      [hl+],a
        ld      a,c    
        ld      [hl+],a
       
        ; bc still points to context->next which just so happens to be where we want it..
        ;
        
        ld      h,b
        ld      l,c

        ; hl now points at our new context

	inc 	hl
	inc	hl
        
	; hl now points to the next contexts stackpointer

	; bc = context->stackpointer

        ld      a,[hl+]
        ld      b,a
        ld      a,[hl+]
        ld      c,a 

IF      DEF(USESLEEP)

	; save our new stack pointer
	push	bc

	; de = context->sleeptill_high

        ld      a,[hl+]
        ld      d,a
        ld      a,[hl+]
        ld      e,a

	; bc = clock_high

	ld	hl,clock_high
        ld      a,[hl+]
        ld      b,a
        ld      a,[hl+]
        ld      c,a

	; compare high.high to high.high

	ld	a,b
	cp	d

	; if myclock_hh > systemclock_hh then 
	; obviously I must be curently sleeping..

	jr	c, .dosleep

	; and if myclock_hh != systemclock_hh
	; then myclock_hh < systemclock_hh
	; so no we are certainly not sleeping

	jr	nz, .nosleep 

	; compare high.low to high.low

	ld	a,c
	cp	e

	; if myclock_hl > systemclock_hl then 
	; obviously I must be curently sleeping..
	; because we already know that
	; myclock_hh = systemclock_hh

	jr	c, .dosleep

	; and if myclock_hl != systemclock_hl
	; then myclock_hl < systemclock_hl
	; so no we are certainly not sleeping

	jr	nz, .nosleep 

	; so myclock_high = systemclock_high - damn this is close
	; its down to the last 16 bits

	; bc = clock_low

	;ld	hl,clock_low
        ld      a,[hl+]
        ld      b,a
        ld      a,[hl+]
        ld      c,a

	; de = context->myclock_low

	; retrieve our context

        ld      hl,context
        ld      a,[hl+]
        ld      d,a
        ld      a,[hl+]
	ld	e,a

	ld	h,d
	ld	l,e

	xor	a
	ld	d,a
	ld	e,context_sleeptill_low	
	add	hl,de	

        ld      a,[hl+]
        ld      d,a
        ld      a,[hl+]
        ld      e,a

	; compare low.high to low.high

	ld	a,b
	cp	d

	; if myclock_lh > systemclock_lh then 
	; obviously I must be curently sleeping..

	jr	c, .dosleep

	; and if myclock_lh != systemclock_lh
	; then myclock_lh < systemclock_lh
	; so no we are certainly not sleeping

	jr	nz, .nosleep 

	; compare low.low to low.low

	ld	a,c
	cp	e

	; if myclock_ll > systemclock_ll then 
	; obviously I must be curently sleeping..
	; because we already know that
	; myclock_lh = systemclock_lh

	jr	c, .dosleep

	; and if myclock_ll != systemclock_ll
	; then myclock_ll < systemclock_ll
	; so no we are certainly not sleeping

	jr	nz, .nosleep 

	; so myclock_low = systemclock_low - its a dead heat

	; So lets let it through..

	jr	.dosleep

.nosleep

	; bc = stackpointer we saved before
	pop	bc
ENDC

        ld      h,b
        ld      l,c

	; repoint our sp

	ld	sp,hl

	;
	; Should do a BANK check on pc=(sp+10)
	; and switch banks if needed.
	; next version perhaps..

	; and the rest follows naturally

	; retrieve old contexts register contents

        pop	hl
	pop	de
	pop	bc
	pop	af

	; and pop out as if this interupt had just finsihed for the new context :)
        reti  
IF      DEF(USESLEEP)
.dosleep

	; discard our new stack pointer
	pop	bc

	; context = context->next
             
        ld      hl,context
        ld      a,[hl+]
        ld      b,a
        ld      a,[hl]
        ld      c,a
        
        ; copy to hl
        ld      h,b
        ld      l,c
        
        ;hl = bc
        
        ; hl now points at our current context of which the first two bytes is
        ; the address of the next context in this chain
        
        ; bc = *context.next

        ld      a,[hl+]
        ld      b,a
        ld      a,[hl+]
        ld      c,a


	jr	.switch	
ENDC

irq_HiLo::
        ei
        reti

SECTION        "InterruptVars",BSS

EXPORT context
tempsp:	     DS	2
context:     DS  2
ringhead:    DS  2
ringtail:    DS  2
clock_high:  DS  2
clock_low:   DS  2
tmodulus:    DS  1


;* End of File *
